@@ -1,10 +1,10 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2002
+ * Copyright (c) 1996-2003
  *	Sleepycat Software.  All rights reserved.
  *
- * $Id: mutex.h,v 11.71 2002/09/10 01:36:48 bostic Exp $
+ * $Id: mutex.h,v 11.90 2003/09/20 21:40:49 bostic Exp $
  */
 
 #ifndef _DB_MUTEX_H_
@@ -77,7 +77,10 @@
 #ifdef HAVE_MUTEX_AIX_CHECK_LOCK
 #include <sys/atomic_op.h>
 typedef int tsl_t;
+
+#ifndef	MUTEX_ALIGN
 #define	MUTEX_ALIGN	sizeof(int)
+#endif
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
 #define	MUTEX_INIT(x)	0
@@ -86,6 +89,25 @@ typedef int tsl_t;
 #endif
 #endif
 
+/*********************************************************************
+ * Apple/Darwin library functions.
+ *********************************************************************/
+#ifdef HAVE_MUTEX_DARWIN_SPIN_LOCK_TRY
+typedef u_int32_t tsl_t;
+
+#ifndef	MUTEX_ALIGN
+#define	MUTEX_ALIGN	sizeof(int)
+#endif
+
+#ifdef LOAD_ACTUAL_MUTEX_CODE
+extern int _spin_lock_try(tsl_t *);
+extern void _spin_unlock(tsl_t *);
+#define	MUTEX_SET(tsl)          _spin_lock_try(tsl)
+#define	MUTEX_UNSET(tsl)        _spin_unlock(tsl)
+#define	MUTEX_INIT(tsl)         (MUTEX_UNSET(tsl), 0)
+#endif
+#endif
+
 /*********************************************************************
  * General C library functions (msemaphore).
  *
@@ -104,14 +126,16 @@ typedef int tsl_t;
 #define	MUTEX_NO_MALLOC_LOCKS
 #define	MUTEX_NO_SHMGET_LOCKS
 
+#ifndef	MUTEX_ALIGN
 #define	MUTEX_ALIGN	16
 #endif
+#endif
 
 #if defined(HAVE_MUTEX_MSEM_INIT) || defined(HAVE_MUTEX_HPPA_MSEM_INIT)
 #include <sys/mman.h>
 typedef msemaphore tsl_t;
 
-#ifndef MUTEX_ALIGN
+#ifndef	MUTEX_ALIGN
 #define	MUTEX_ALIGN	sizeof(int)
 #endif
 
@@ -128,7 +152,9 @@ typedef msemaphore tsl_t;
 #ifdef HAVE_MUTEX_PLAN9
 typedef Lock tsl_t;
 
+#ifndef	MUTEX_ALIGN
 #define	MUTEX_ALIGN	sizeof(int)
+#endif
 
 #define	MUTEX_INIT(x)	(memset(x, 0, sizeof(Lock)), 0)
 #define	MUTEX_SET(x)	canlock(x)
@@ -159,7 +185,10 @@ typedef spinlock_t tsl_t;
 #ifdef HAVE_MUTEX_SEMA_INIT
 #include <synch.h>
 typedef sema_t tsl_t;
+
+#ifndef	MUTEX_ALIGN
 #define	MUTEX_ALIGN	 sizeof(int)
+#endif
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
 #define	MUTEX_DESTROY(x) sema_destroy(x)
@@ -175,7 +204,10 @@ typedef sema_t tsl_t;
 #ifdef HAVE_MUTEX_SGI_INIT_LOCK
 #include <abi_mutex.h>
 typedef abilock_t tsl_t;
+
+#ifndef	MUTEX_ALIGN
 #define	MUTEX_ALIGN	sizeof(int)
+#endif
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
 #define	MUTEX_INIT(x)	(init_lock(x) != 0)
@@ -194,7 +226,10 @@ typedef abilock_t tsl_t;
 #ifdef HAVE_MUTEX_SOLARIS_LOCK_TRY
 #include <sys/machlock.h>
 typedef lock_t tsl_t;
+
+#ifndef	MUTEX_ALIGN
 #define	MUTEX_ALIGN	sizeof(int)
+#endif
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
 #define	MUTEX_INIT(x)	0
@@ -209,8 +244,11 @@ typedef lock_t tsl_t;
 #ifdef HAVE_MUTEX_VMS
 #include <sys/mman.h>;
 #include <builtins.h>
-typedef unsigned char tsl_t;
+typedef volatile unsigned char tsl_t;
+
+#ifndef	MUTEX_ALIGN
 #define	MUTEX_ALIGN		sizeof(unsigned int)
+#endif
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
 #ifdef __ALPHA
@@ -234,7 +272,10 @@ typedef unsigned char tsl_t;
 #ifdef HAVE_MUTEX_VXWORKS
 #include "taskLib.h"
 typedef SEM_ID tsl_t;
+
+#ifndef	MUTEX_ALIGN
 #define	MUTEX_ALIGN		sizeof(unsigned int)
+#endif
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
 #define	MUTEX_SET(tsl)		(semTake((*tsl), WAIT_FOREVER) == OK)
@@ -283,7 +324,10 @@ do {									\
  *********************************************************************/
 #ifdef HAVE_MUTEX_WIN16
 typedef unsigned int tsl_t;
+
+#ifndef	MUTEX_ALIGN
 #define	MUTEX_ALIGN		sizeof(unsigned int)
+#endif
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
 #define	MUTEX_INIT(x)		0
@@ -295,7 +339,7 @@ typedef unsigned int tsl_t;
 /*********************************************************************
  * Win32
  *********************************************************************/
-#ifdef HAVE_MUTEX_WIN32
+#if defined(HAVE_MUTEX_WIN32) || defined(HAVE_MUTEX_WIN32_GCC)
 #define	MUTEX_FIELDS							\
 	LONG tas;							\
 	LONG nwaiters;							\
@@ -305,6 +349,22 @@ typedef unsigned int tsl_t;
 #define	MUTEX_SET(tsl)		(!InterlockedExchange((PLONG)tsl, 1))
 #define	MUTEX_UNSET(tsl)	(*(tsl) = 0)
 #define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
+
+/*
+ * From Intel's performance tuning documentation (and see SR #6975):
+ * ftp://download.intel.com/design/perftool/cbts/appnotes/sse2/w_spinlock.pdf
+ *
+ * "For this reason, it is highly recommended that you insert the PAUSE
+ * instruction into all spin-wait code immediately. Using the PAUSE
+ * instruction does not affect the correctness of programs on existing
+ * platforms, and it improves performance on Pentium 4 processor platforms."
+ */
+#ifdef HAVE_MUTEX_WIN32
+#define	MUTEX_PAUSE		{__asm{_emit 0xf3}; __asm{_emit 0x90}}
+#endif
+#ifdef HAVE_MUTEX_WIN32_GCC
+#define	MUTEX_PAUSE		asm volatile ("rep; nop" : : );
+#endif
 #endif
 #endif
 
@@ -339,14 +399,16 @@ typedef unsigned char tsl_t;
  *********************************************************************/
 #ifdef HAVE_MUTEX_ALPHA_GCC_ASSEMBLY
 typedef u_int32_t tsl_t;
+
+#ifndef	MUTEX_ALIGN
 #define	MUTEX_ALIGN	4
+#endif
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
 /*
  * For gcc/alpha.  Should return 0 if could not acquire the lock, 1 if
  * lock was acquired properly.
  */
-#ifdef __GNUC__
 static inline int
 MUTEX_SET(tsl_t *tsl) {
 	register tsl_t *__l = tsl;
@@ -374,18 +436,30 @@ MUTEX_UNSET(tsl_t *tsl) {
 	asm volatile("	mb\n");
 	return *tsl = 0;
 }
-#endif
-
-#ifdef __DECC
-#include <alpha/builtins.h>
-#define	MUTEX_SET(tsl)		(__LOCK_LONG_RETRY((tsl), 1) != 0)
-#define	MUTEX_UNSET(tsl)	(*(tsl) = 0)
-#endif
 
 #define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
 #endif
 #endif
 
+/*********************************************************************
+ * Tru64/cc assembly.
+ *********************************************************************/
+#ifdef HAVE_MUTEX_TRU64_CC_ASSEMBLY
+typedef volatile u_int32_t tsl_t;
+
+#ifndef	MUTEX_ALIGN
+#define	MUTEX_ALIGN	4
+#endif
+
+#ifdef LOAD_ACTUAL_MUTEX_CODE
+#include <alpha/builtins.h>
+#define	MUTEX_SET(tsl)		(__LOCK_LONG_RETRY((tsl), 1) != 0)
+#define	MUTEX_UNSET(tsl)	(__UNLOCK_LONG(tsl))
+
+#define	MUTEX_INIT(tsl)		(MUTEX_UNSET(tsl), 0)
+#endif
+#endif
+
 /*********************************************************************
  * ARM/gcc assembly.
  *********************************************************************/
@@ -398,10 +472,11 @@ typedef unsigned char tsl_t;
  */
 #define	MUTEX_SET(tsl) ({						\
 	int __r;							\
-	asm volatile("swpb %0, %1, [%2]"				\
-	    : "=r" (__r) 						\
-	    : "0" (1), "r" (tsl)					\
-	    : "memory"							\
+	asm volatile(							\
+		"swpb	%0, %1, [%2]\n\t"				\
+		"eor	%0, %0, #1\n\t"					\
+	    : "=&r" (__r)						\
+	    : "r" (1), "r" (tsl)					\
 	    );								\
 	__r & 1;							\
 })
@@ -411,12 +486,55 @@ typedef unsigned char tsl_t;
 #endif
 #endif
 
+/*********************************************************************
+ * MIPS/gcc assembly.
+ *********************************************************************/
+#ifdef HAVE_MUTEX_MIPS_GCC_ASSEMBLY
+typedef u_int32_t tsl_t;
+
+#ifndef        MUTEX_ALIGN
+#define        MUTEX_ALIGN     4
+#endif
+
+#ifdef LOAD_ACTUAL_MUTEX_CODE
+/*
+ * For gcc/mips.  Should return 0 if could not acquire the lock, 1 if
+ * lock was acquired properly.
+ */
+static inline int
+MUTEX_SET(tsl_t *tsl) {
+       register tsl_t *__l = tsl;
+       register tsl_t __r;
+       __asm__ __volatile__(
+               "       .set push       \n"
+               "       .set mips2      \n"
+               "       .set noreorder  \n"
+               "       .set nomacro    \n"
+               "1:     ll      %0, %1          \n"
+               "       bne     %0, $0, 1f      \n"
+               "        xori   %0, %0, 1       \n"
+               "       sc      %0, %1          \n"
+               "       beql    %0, $0, 1b      \n"
+               "        xori   %0, 1           \n"
+               "1:     .set pop                  "
+               : "=&r" (__r), "+R" (*__l));
+       return __r;
+}
+
+#define        MUTEX_UNSET(tsl)        (*(volatile tsl_t *)(tsl) = 0)
+#define        MUTEX_INIT(tsl)         MUTEX_UNSET(tsl)
+#endif
+#endif
+
 /*********************************************************************
  * HPPA/gcc assembly.
  *********************************************************************/
 #ifdef HAVE_MUTEX_HPPA_GCC_ASSEMBLY
 typedef u_int32_t tsl_t;
+
+#ifndef	MUTEX_ALIGN
 #define	MUTEX_ALIGN	16
+#endif
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
 /*
@@ -465,8 +583,7 @@ typedef unsigned char tsl_t;
 /*********************************************************************
  * PowerPC/gcc assembly.
  *********************************************************************/
-#if defined(HAVE_MUTEX_PPC_GENERIC_GCC_ASSEMBLY) ||			\
-    (HAVE_MUTEX_PPC_APPLE_GCC_ASSEMBLY)
+#if defined(HAVE_MUTEX_PPC_GCC_ASSEMBLY)
 typedef u_int32_t tsl_t;
 
 #ifdef LOAD_ACTUAL_MUTEX_CODE
@@ -500,54 +617,54 @@ typedef u_int32_t tsl_t;
  *
  * 'set' mutexes have the value 1, like on Intel; the returned value from
  * MUTEX_SET() is 1 if the mutex previously had its low bit clear, 0 otherwise.
- *
- * Mutexes on Mac OS X work the same way as the standard PowerPC version, but
- * the assembler syntax is subtly different -- the standard PowerPC version
- * assembles but doesn't work correctly.  This version makes (unnecessary?)
- * use of a stupid linker trick: __db_mutex_tas_dummy is never called, but the
- * ___db_mutex_set label is used as a function name.
  */
-#ifdef HAVE_MUTEX_PPC_APPLE_GCC_ASSEMBLY
-extern int __db_mutex_set __P((volatile tsl_t *));
-void
-__db_mutex_tas_dummy()
-{
-	__asm__	__volatile__("		\n\
-	.globl 	___db_mutex_set		\n\
-___db_mutex_set:			\n\
-	lwarx	r5,0,r3			\n\
-	cmpwi 	r5,0			\n\
-	bne 	fail			\n\
-	addi 	r5,r5,1			\n\
-	stwcx. 	r5,0,r3			\n\
-	beq 	success			\n\
-fail:					\n\
-	li 	r3,0			\n\
-	blr 				\n\
-success:				\n\
-	li 	r3,1			\n\
-	blr");	
+#ifdef HAVE_MUTEX_PPC_GCC_ASSEMBLY
+static inline int
+MUTEX_SET(int *tsl)  {
+         int __r;
+         asm volatile (
+"0:                             \n\t"
+"       lwarx   %0,0,%1         \n\t"
+"       cmpwi   %0,0            \n\t"
+"       bne-    1f              \n\t"
+"       stwcx.  %1,0,%1         \n\t"
+"       isync                   \n\t"
+"       beq+    2f              \n\t"
+"       b       0b              \n\t"
+"1:                             \n\t"
+"       li      %1, 0           \n\t"
+"2:                             \n\t"
+         : "=&r" (__r), "+r" (tsl)
+         :
+         : "cr0", "memory");
+         return (int)tsl;
 }
-#define	MUTEX_SET(tsl)  __db_mutex_set(tsl)
 #endif
-#ifdef HAVE_MUTEX_PPC_GENERIC_GCC_ASSEMBLY
-#define	MUTEX_SET(tsl)	({		\
-	int __one = 1;			\
-	int __r;			\
-	tsl_t *__l = (tsl);		\
-	asm volatile ("			\
-0:					\
-	lwarx %0,0,%1;			\
-	cmpwi %0,0;			\
-	bne 1f;				\
-	stwcx. %2,0,%1;			\
-	bne- 0b;			\
-1:"					\
-	: "=&r" (__r)			\
-	: "r" (__l), "r" (__one));	\
-	!(__r & 1);			\
-})
+static inline int
+MUTEX_UNSET(tsl_t *tsl) {
+         asm volatile("sync" : : : "memory");
+         return *tsl = 0;
+}
+#define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
 #endif
+#endif
+
+/*********************************************************************
+ * OS/390 C
+ *********************************************************************/
+#ifdef HAVE_MUTEX_S390_CC_ASSEMBLY
+typedef int tsl_t;
+
+#ifndef	MUTEX_ALIGN
+#define	MUTEX_ALIGN	 sizeof(int)
+#endif
+
+#ifdef LOAD_ACTUAL_MUTEX_CODE
+/*
+ * cs() is declared in <stdlib.h> but is built in to the compiler.
+ * Must use LANGLVL(EXTENDED) to get its declaration.
+ */
+#define	MUTEX_SET(tsl)		(!cs(&zero, (tsl), 1))
 #define	MUTEX_UNSET(tsl)	(*(tsl) = 0)
 #define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
 #endif
@@ -657,7 +774,10 @@ typedef unsigned char tsl_t;
 #ifdef HAVE_MUTEX_UTS_CC_ASSEMBLY
 typedef int tsl_t;
 
+#ifndef	MUTEX_ALIGN
 #define	MUTEX_ALIGN	sizeof(int)
+#endif
+
 #ifdef LOAD_ACTUAL_MUTEX_CODE
 #define	MUTEX_INIT(x)	0
 #define	MUTEX_SET(x)	(!uts_lock(x, 1))
@@ -687,6 +807,17 @@ typedef unsigned char tsl_t;
 
 #define	MUTEX_UNSET(tsl)	(*(tsl) = 0)
 #define	MUTEX_INIT(tsl)		MUTEX_UNSET(tsl)
+
+/*
+ * From Intel's performance tuning documentation (and see SR #6975):
+ * ftp://download.intel.com/design/perftool/cbts/appnotes/sse2/w_spinlock.pdf
+ *
+ * "For this reason, it is highly recommended that you insert the PAUSE
+ * instruction into all spin-wait code immediately. Using the PAUSE
+ * instruction does not affect the correctness of programs on existing
+ * platforms, and it improves performance on Pentium 4 processor platforms."
+ */
+#define	MUTEX_PAUSE		asm volatile ("rep; nop" : : );
 #endif
 #endif
 
@@ -714,33 +845,38 @@ typedef unsigned char tsl_t;
 
 /*
  * !!!
- * These defines are separated into the u_int8_t flags stored in the
- * mutex below, and the 32 bit flags passed to __db_mutex_setup.
- * But they must co-exist and not overlap.  Flags to __db_mutex_setup are:
+ * The flag arguments for __db_mutex_setup (and the underyling intialization
+ * function for the mutex type, for example, __db_tas_mutex_init), and flags
+ * stored in the DB_MUTEX structure are combined, and may not overlap.  Flags
+ * to __db_mutex_setup:
  *
- * MUTEX_ALLOC - Use when the mutex to initialize needs to be allocated.
- *    The 'ptr' arg to __db_mutex_setup should be a DB_MUTEX ** whenever
- *    you use this flag.  If this flag is not set, the 'ptr' arg is
- *    a DB_MUTEX *.
- * MUTEX_NO_RECORD - Explicitly do not record the mutex in the region.
- *    Otherwise the mutex will be recorded by default.  If you set
- *    this you need to understand why you don't need it recorded.  The
- *    *only* ones not recorded are those that are part of region structures
- *    that only get destroyed when the regions are destroyed.
- * MUTEX_NO_RLOCK - Explicitly do not lock the given region otherwise
- *    the region will be locked by default.
- * MUTEX_SELF_BLOCK - Set if self blocking mutex.
- * MUTEX_THREAD - Set if mutex is a thread-only mutex.
+ * MUTEX_ALLOC:
+ *	Use when the mutex to initialize needs to be allocated. The 'ptr'
+ *	arg to __db_mutex_setup should be a DB_MUTEX ** whenever you use
+ *	this flag.  If this flag is not set, the 'ptr' arg is a DB_MUTEX *.
+ * MUTEX_NO_RECORD:
+ *	Explicitly do not record the mutex in the region.  Otherwise the
+ *	mutex will be recorded by default.  If you set this you need to
+ *	understand why you don't need it recorded.  The *only* ones not
+ *	recorded are those that are part of region structures that only
+ *	get destroyed when the regions are destroyed.
+ * MUTEX_NO_RLOCK:
+ *	Explicitly do not lock the given region otherwise the region will
+ *	be locked by default.
+ * MUTEX_SELF_BLOCK:
+ *	Set if self blocking mutex.
+ * MUTEX_THREAD:
+ *	Set if mutex is a thread-only mutex.
  */
-#define	MUTEX_IGNORE		0x001	/* Ignore, no lock required. */
-#define	MUTEX_INITED		0x002	/* Mutex is successfully initialized */
-#define	MUTEX_MPOOL		0x004	/* Allocated from mpool. */
-#define	MUTEX_SELF_BLOCK	0x008	/* Must block self. */
-/* Flags only, may be larger than 0xff. */
-#define	MUTEX_ALLOC		0x00000100 /* Allocate and init a mutex */
-#define	MUTEX_NO_RECORD		0x00000200 /* Do not record lock */
-#define	MUTEX_NO_RLOCK		0x00000400 /* Do not acquire region lock */
-#define	MUTEX_THREAD		0x00000800 /* Thread-only mutex. */
+#define	MUTEX_ALLOC		0x0001	/* Allocate and init a mutex */
+#define	MUTEX_IGNORE		0x0002	/* Ignore, no lock required. */
+#define	MUTEX_INITED		0x0004	/* Mutex is successfully initialized */
+#define	MUTEX_LOGICAL_LOCK	0x0008	/* Mutex backs database lock. */
+#define	MUTEX_MPOOL		0x0010	/* Allocated from mpool. */
+#define	MUTEX_NO_RECORD		0x0020	/* Do not record lock */
+#define	MUTEX_NO_RLOCK		0x0040	/* Do not acquire region lock */
+#define	MUTEX_SELF_BLOCK	0x0080	/* Must block self. */
+#define	MUTEX_THREAD		0x0100	/* Thread-only mutex. */
 
 /* Mutex. */
 struct __mutex_t {
@@ -750,7 +886,6 @@ struct __mutex_t {
 #else
 	tsl_t	tas;			/* Test and set. */
 #endif
-	u_int32_t spins;		/* Spins before block. */
 	u_int32_t locked;		/* !0 if locked. */
 #else
 	u_int32_t off;			/* Byte offset to lock. */
@@ -763,8 +898,12 @@ struct __mutex_t {
 #ifdef HAVE_MUTEX_SYSTEM_RESOURCES
 	roff_t	  reg_off;		/* Shared lock info offset. */
 #endif
-
-	u_int8_t  flags;		/* MUTEX_XXX */
+	/*
+	 * Flags should be an unsigned integer even if it's not required by
+	 * the possible flags values, getting a single byte on some machines
+	 * is expensive, and the mutex structure is a MP hot spot.
+	 */
+	u_int32_t flags;		/* MUTEX_XXX */
 };
 
 /* Redirect calls to the correct functions. */
@@ -776,7 +915,7 @@ struct __mutex_t {
 #define	__db_mutex_lock(a, b)		__db_pthread_mutex_lock(a, b)
 #define	__db_mutex_unlock(a, b)		__db_pthread_mutex_unlock(a, b)
 #define	__db_mutex_destroy(a)		__db_pthread_mutex_destroy(a)
-#elif defined(HAVE_MUTEX_WIN32)
+#elif defined(HAVE_MUTEX_WIN32) || defined(HAVE_MUTEX_WIN32_GCC)
 #define	__db_mutex_init_int(a, b, c, d)	__db_win32_mutex_init(a, b, d)
 #define	__db_mutex_lock(a, b)		__db_win32_mutex_lock(a, b)
 #define	__db_mutex_unlock(a, b)		__db_win32_mutex_unlock(a, b)
